#if !defined  HAVE_SLOWCNVL_H__
#define       HAVE_SLOWCNVL_H__

#include "fxttypes.h"

#include "copy.h"
#define copy fxtaux::copy // avoid unwanted std::copy

template <typename Type>
void slow_convolution(const Type *f, const Type *g, Type *h, ulong n)
// (cyclic) convolution:  h[] :=  f[] (*) g[]
// n := array length
{
#if 0
    for (ulong tau=0;  tau<n;  ++tau)
    {
        Type s = 0.0;
        for (ulong k=0; k<n; ++k)
        {
            ulong k2 = tau - k;
            if ( (long)k2<0 )  k2 += n;
            s += (f[k]*g[k2]);
        }
        h[tau] = s;
    }
#else // 0
    // avoid if in inner loop:
    for (ulong tau=0;  tau<n;  ++tau)
    {
        Type s = 0.0;
        ulong k = 0;
        for (ulong k2=tau;  k<=tau;  ++k, --k2)  s += (f[k]*g[k2]);
        for (ulong k2=n-1;  k<n;     ++k, --k2)  s += (f[k]*g[k2]);
        h[tau] = s;
    }
#endif // 0
}
// -------------------------


template <typename Type>
void slow_convolution0(const Type *f, const Type *g, Type *h, ulong n)
// (linear) convolution:  h[] :=  f[] (*0) g[]
// n := array length
//
// input data must be zero padded:
//   f[n/2] .. f[n-1] == 0 and g[n/2] .. g[n-1] == 0
{
    const ulong nh = n/2;
    for (ulong tau=0; tau<nh; ++tau)
    {
        Type s = 0.0;
        for (ulong k=0, k2=tau;  (long)k2>=0;  ++k, --k2)  s += (f[k]*g[k2]);
        h[tau] = s;
    }

    for (ulong tau=nh; tau<n; ++tau)
    {
        Type s = 0.0;
        for (ulong k=tau-nh, k2=nh;  (long)k2>=0;  ++k, --k2)  s += (f[k]*g[k2]);
        h[tau] = s;
    }
}
// -------------------------


template <typename Type>
void slow_auto_convolution(const Type *f, Type *g, ulong n)
// (cyclic) convolution:  g[] :=  f[] (*) g[]
// n := array length
{
    slow_convolution(f, f, g, n);
}
// -------------------------

template <typename Type>
void slow_auto_convolution0(const Type *f, Type *g, ulong n)
// (linear) self convolution:  g[] :=  f[] (*0) g[]
// n := array length
//
// input data must be zero padded:
//   f[n/2] .. f[n-1] == 0 and g[n/2] .. g[n-1] == 0
// n  must be >=2
{
    slow_convolution0(f, f, g, n);
}
// -------------------------


template <typename Type>
void slow_convolution(const Type *f, Type *g, ulong n)
// (cyclic) convolution:  g[] :=  f[] (*) g[]
// n := array length
{
    Type h[n];
    slow_convolution(f, g, h, n);
    copy(h, g, n);
}
// -------------------------

template <typename Type>
void slow_convolution0(const Type *f, Type *g, ulong n)
// (linear) convolution:  g[] :=  f[] (*0) g[]
// n := array length
//
// input data must be zero padded:
//   f[n/2] .. f[n-1] == 0 and g[n/2] .. g[n-1] == 0
// n  must be >=2
{
    Type h[n];
    slow_convolution0(f, g, h, n);
    copy(h, g, n);
}
// -------------------------


template <typename Type>
void slow_auto_convolution(Type *f, ulong n)
// (cyclic) self convolution:  f[] :=  f[] (*) f[]
// n := array length
{
    Type h[n];
    slow_convolution(f, f, h, n);
    copy(h, f, n);
}
// -------------------------

template <typename Type>
void slow_auto_convolution0(Type *f, ulong n)
// (linear) self convolution:  g[] :=  f[] (*0) g[]
// n := array length
//
// input data must be zero padded:
//   f[n/2] .. f[n-1] == 0
// n  must be >=2
{
    Type h[n];
    slow_convolution0(f, f, h, n);
    copy(h, f, n);
}
// -------------------------

#undef copy  // avoid unwanted std::copy


#endif  // !defined HAVE_SLOWCNVL_H__
